# Utility functions and classes

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys,os,threading,string,cPickle,types,time,traceback,pprint,socket,random,Queue

from error import Error
import settings

# GTK utilities

# Some complexity is added to work under win32
# You may scream in horror if you wish.
# Possibly should migrate to a better cross platform library one day
#   QT?

threadnice_mainloop_status = 'running'
action_timeout_added = 0

def threadnice_mainloop():
    """ Like gtk.mainloop.
            Work around a bug in timeout_add under win32 """
    global threadnice_mainloop_status
    import gtk

    if sys.platform == 'win32':
        while threadnice_mainloop_status == 'running':
            _gtk_action_timeout()
            while gtk.events_pending() and threadnice_mainloop_status == 'running':
                gtk.mainiteration()
            time.sleep(0.05)
        threadnice_mainloop_status = 'running'

    else:
        gtk.mainloop()

def threadnice_mainquit(*args):
    """ Like gtk.mainquit
            Work around a bug in timeout_add under win32 """
    global threadnice_mainloop_status, action_timeout_added

    #if sys.platform == 'win32':
    if threadnice_mainloop_status == 'fake' or sys.platform == 'win32':
        threadnice_mainloop_status = 'stopped'
    else:
        import gtk
        gtk.mainquit()

# Allow faking a mainloop
# ie to make schedule_gtk, do_gtk still work
# This is used in the text mode interface

def threadnice_mainloop_fake_start():
    global threadnice_mainloop_status
    threadnice_mainloop_status = 'fake'
def threadnice_mainloop_fake_stop():
    global threadnice_mainloop_running
    threadnice_mainloop_running = 'stopped'
def threadnice_mainloop_fake_iteration():
    _gtk_action_timeout()
def threadnice_mainloop_fake_is_running():
    return threadnice_mainloop_status == 'fake'

# Provide a mechanism for
#  - running code in the main thread (for use with GTK, which is not threadsafe)
#  - running code after an interval

gtk_action_lock = threading.RLock()
gtk_action_list = [ ]

def _gtk_action_timeout():
    """ (internal) Check for and run any pending actions. """

    while 1:
        gtk_action_lock.acquire()
        try:
            if not gtk_action_list or gtk_action_list[0][0] > time.time():
                break
            head = gtk_action_list.pop(0)
        finally:
            gtk_action_lock.release()

        try:
            if apply(head[1],head[2]):
                apply(schedule_gtk,(head[3],head[1])+head[2])
        except:
            traceback.print_exc()
                
    return 1

def _ensure_action_timeout():
    """ (internal) Ensure timeout_add has been called. """
    global action_timeout_added

    if sys.modules.has_key('gtk') and sys.platform <> 'win32' and not action_timeout_added:
        import gtk
        gtk.timeout_add(100, _gtk_action_timeout)
        action_timeout_added = 1

def schedule_gtk(t_plus,func,*parameters):
    """ Schedule an action: 
                after t_plus milliseconds, func will be called in the main thread with
                the parameters given. 
                
            If func returns true, it will be called again after another t_plus
            milliseconds"""
            
    _ensure_action_timeout()

    gtk_action_lock.acquire()
    try:
        item = (time.time()+t_plus/1000.0,func,parameters,t_plus)
        pos = 0
        while pos < len(gtk_action_list) and gtk_action_list[pos][0] <= item[0]:
            pos = pos + 1
        gtk_action_list.insert(pos, item)
    finally:
        gtk_action_lock.release()

def unschedule_gtk(func):
    """ Remove all pending func actions. """

    global gtk_action_list
    gtk_action_lock.acquire()
    try:
        gtk_action_list = filter(lambda x,func=func: x <> func,gtk_action_list)
    finally:
        gtk_action_lock.release()


def do_gtk(func,*parameters):
    """ Do an action as soon as possible.

            This is useful primarily to interact with gtk. """

    apply(schedule_gtk,(0.0,func)+parameters)

open_windows = [ ]

def show_window(window, name, close_on_exit = 1, show_all = 1):
    """ Open a window. Add to list of open windows, so that all windows
            can be closed on exit. """

    pair = (window, close_on_exit)

    if pair not in open_windows:
        window.set_wmclass(name,'the_circle')
        open_windows.append((window,close_on_exit))
        window.connect("destroy", lambda window, pair=pair: open_windows.remove(pair))

    if show_all:
        window.show_all()
    else:
        window.show()

def close_all_windows():
    """ Close all open windows opened with show_window and the close_on_exit option
            set to true. """

    for pair in open_windows[:]:
        if pair[1]:
            pair[0].destroy()

# Win32 GtkTooltips have problems
if sys.platform == 'win32':
    def _checker():
        any = 0
        for pair in open_windows[:]:
            point = pair[0].get_pointer()
            if point[0] < 0 or point[1] < 0:
                continue
            alloc = pair[0].get_allocation()
            if point[0] >= alloc[2] or point[1] >= alloc[3]:
                continue
            any = 1
            break

        from widgets import tooltips
        if any:
            tooltips.enable()
        else:
            tooltips.disable()
            
        return 1

    schedule_gtk(2000, _checker)

# Lists

def split_list(str):
    """ Split a comma or space separated string. """

    return string.split(string.replace(str,',',' '))

# Alerts

def beep():
    """ Produce a beep. """

    if sys.platform == 'win32':
        import winsound
        winsound.PlaySound('SystemAsterisk',winsound.SND_ALIAS | winsound.SND_ASYNC)
    elif sys.modules.has_key('gtk'):
        import gtk
        gtk.gdk_beep()
    else:
        sys.stdout.write(chr(7))
        sys.stdout.flush()

# DNS queries

def hostname_of_ip(ip):
    """ Perform a DNS lookup on an IP address. """

    try:
        return socket.gethostbyaddr(ip)[0]
    except socket.error:
        return ip

# URL viewing

original_directory = os.getcwd()

def find_file(file):
    possibilities = [ os.path.abspath(file) ]
    for item in sys.path:
        possibilities.append(os.path.abspath(os.path.join(original_directory,item,'circlelib',file)))

    for item in possibilities:
        if os.path.isfile(item):
            break
    else:
        raise Error('File does not exist.')

    return item

def browse_url(url):
    if sys.platform == 'win32':
        try:
            import webbrowser
            webbrowser.open_new(url)
        except:
            raise Error('Opening a browser failed.')
    elif sys.modules.has_key('gtk'):
        try:
            import gnome.url
        except:
            gnome = None
        if gnome:
            gnome.url.show(url)
        else:
            try:
                if not os.fork():
                    os.execlp('gnome-moz-remote','gnome-moz-remote','--newwin',url)
            except:
                raise Error('Opening a browser failed.')
    else:
        try:
            #if not os.fork():
                #os.execlp('lynx','lynx',url)
            os.system("lynx "+quote_for_shell(url))
        except:
            raise Error('Opening a browser failed.')


def browse_file(file,section=''):
    browse_url('file://'+find_file(file)+section)

# pydoc

pydoc_url = None

def browse_modules():
    class Server(threading.Thread):
        def run(self):
            global pydoc_url
            try:
                import pydoc
            except:
                pydoc_url = -1
                return

            def ready(server):
                global pydoc_url
                pydoc_url = server.url
            for i in xrange(1024,65535):
                try:
                    pydoc.serve(i,ready)
                    break
                except:
                    pass
            else:
                pydoc_url = -1

    if not pydoc_url:
        server = Server()
        server.setDaemon(1)
        server.start()

    while not pydoc_url:
        time.sleep(0.5)
                
    if not type(pydoc_url) == type(''):
        raise Error("Could not start pydoc server.")

    browse_url(pydoc_url+'circlelib') 

# Status windows

class Status:
    def show(self, title,message, window=None, icon=None):
        import gtk

        self.window = gtk.GtkWindow()
        self.window.set_border_width(0)        
        self.window.set_default_size(300,10)
        if window:
            self.window.set_transient_for(window)
        self.window.set_title(title)

        hbox = gtk.GtkHBox(0, 0)
        self.window.add(hbox)

        if icon:
            pixmap = gtk.GtkPixmap(self.window,find_file(icon))
            hbox.pack_start(pixmap, 0,0,0)

        align = gtk.GtkAlignment(0,0,0,0)
        hbox.pack_start(align, 0,0,10)

        self.label = gtk.GtkLabel('\n' + message + '\n\n')
        self.label.set_justify(gtk.JUSTIFY_LEFT)
        align.add(self.label)

        show_window(self.window, 'status')

    def close(self):
        self.window.destroy()

    def message(self, message):
        self.label.set_text('\n' + message + '\n\n')


# Synchronized access objects

class Synchronous:
    """ Synchronized access objects"""
    def __init__(self):
        self.lock = threading.RLock()

# Task Management

task_lock = threading.RLock()
task_count = 0
pending_task_queue = Queue.Queue(0)
pending_low_priority_task_queue = Queue.Queue(0)

class Task(threading.Thread):
    def __init__(self, function, manager, *parameters):
        self.function   = function
        self.manager    = manager
        self.parameters = parameters
        
        threading.Thread.__init__(self)

        self.setDaemon(1)

    def __repr__(self):
        if not self.function:
            return '<Task (completed)>'
        else:
            return '<Task '+self.function.__name__+' '+`self.parameters`+'>'

    def start_no_overload(self, low_priority = 0):
        global task_count

        self.include_in_count = 1

        self.manager.lock.acquire()

        #if len(self.manager.tasks) >= max_task_manager_threads:
        #if threading.activeCount() >= max_active_threads:
        if task_count >= settings.max_active_threads:
            self.manager.task_queue.put(1) #(self)
            if low_priority:
                pending_low_priority_task_queue.put(self)
            else:
                pending_task_queue.put(self)
            self.manager.lock.release()
            return
        
        self.manager.tasks.append(self)
        self.manager.lock.release()

        task_lock.acquire()
        task_count = task_count + 1
        task_lock.release()

        threading.Thread.start(self)

    def start(self, include_in_count=1):
        global task_count

        self.include_in_count = include_in_count
        
        self.manager.lock.acquire()
        self.manager.tasks.append(self)
        self.manager.lock.release()

        if include_in_count:
            task_lock.acquire()
            task_count = task_count + 1
            task_lock.release()

        threading.Thread.start(self)

    def run(self):
        global task_count

        try:
            try:
                apply(self.function,(self.manager,) + self.parameters)
            except:
                traceback.print_exc()
        
            if not self.include_in_count:
                return

            self.manager.lock.acquire()
            try:
                self.manager.tasks.remove(self)

                #if threading.activeCount() > max_active_threads:
                if task_count > settings.max_active_threads:
                    task_lock.acquire()
                    task_count = task_count - 1
                    task_lock.release()
                    return

                try:
                    #got = self.manager.task_queue.get(0)
                    got = pending_task_queue.get(0)
                except Queue.Empty:
                    try:
                        got = pending_low_priority_task_queue.get(0)
                    except Queue.Empty:
                        task_lock.acquire()
                        task_count = task_count - 1
                        task_lock.release()
                        return
            finally:
                self.manager.lock.release()

            got.manager.lock.acquire()
            got.manager.tasks.append(got)
            got.manager.task_queue.get(0)
            got.manager.lock.release()

            threading.Thread.start(got)

        finally:
            self.function = None
            self.parameters = None
            self.manager = None

class Task_manager(Synchronous):
    def __init__(self):
        Synchronous.__init__(self)
        self.tasks   = [ ]
        self.task_queue = Queue.Queue(0)
        self.running = 0

    def start(self):
        self.lock.acquire()
        self.running = 1
        self.lock.release()

    def stop(self, fast=0):
        self.lock.acquire()
        self.running = 0
        self.lock.release()

        if fast:
            return

        while 1:
            while not self.task_queue.empty():
                time.sleep(0.5)

            self.lock.acquire()
            if len(self.tasks):
                head = self.tasks[0]
            else:
                head = None
            self.lock.release()

            if not head: 
                break

            head.join()
            
            # In case it died
            self.lock.acquire()
            if head in self.tasks:
                self.tasks.remove(head)
            self.lock.release()

    def start_poller(self, initial_interval,interval, func,*param):
        """ Run a job regularly until the task manager is stopped. """
        def poller(self=self,func=func,param=param,interval=interval):
            self.lock.acquire()
            try:
                if not self.running:
                    return

                if not self.poller_running:
                    self.poller_running = 1

                    def do_it(self,func,param):
                        apply(func,param)
                        self.lock.acquire()
                        self.poller_running = 0
                        self.lock.release()
                    Task(do_it,self,func,param).start() 
            finally:
                self.lock.release()

            unschedule_gtk(self.poller_function)
            schedule_gtk(interval*1000,self.poller_function)
        
        self.poller_function = poller
        self.poller_running = 0
        schedule_gtk(initial_interval*1000,self.poller_function)

    def run_poller_now(self):
        """ Force the poll function to be run now. """
        self.poller_function()

# Pipe task

class Pipe(Task_manager):
    def __init__(self):
        Task_manager.__init__(self)
        
        self.queue = [ ]

    def start(self,function,*parameters):
        Task_manager.start(self)
        apply(Task,(function,self)+parameters).start()

    def finished(self):
        self.lock.acquire()
        try:
            return self.task_queue.empty() and self.tasks == [ ] and self.queue == [ ]
        finally:
            self.lock.release()

    def write(self, item):
        self.lock.acquire()
        self.queue.append(item)
        self.lock.release()

    def read_all(self):
        self.lock.acquire()
        try:
            return self.queue
        finally:
            self.queue = [ ]
            self.lock.release()

    def read_until_finished(self):
        """ Read pipe until finished, then stop pipe. """

        list = [ ]
        while not self.finished():
            list.extend(self.read_all())
            time.sleep(1.0)
        self.stop()

        return list

# Type checking

def check_type(item,the_type):
    if type(item) <> the_type:
        raise Error('data type error')

def check_is_name(item):
    if type(item) <> types.StringType or len(item) <> settings.name_bytes:
        raise Error('data type error')

def check_is_long_name(item):
    if type(item) <> types.StringType or len(item) <> settings.long_name_bytes:
        raise Error('data type error')

def check_is_address(item):
    if type(item) <> types.TupleType or \
         type(item[0]) <> types.StringType or \
         type(item[1]) <> types.IntType:
        raise Error('data type error')

def check_is_public_key(item):
    if type(item) <> types.TupleType or \
         type(item[0]) <> types.LongType or \
         type(item[1]) <> types.LongType:
        raise Error('data type error')

def check(item, template):
    if type(item) <> type(template):
        raise Error('data type error')

    if type(item) in [types.TupleType, types.ListType]:
        if len(item) < len(template):
            raise Error('data type error')
        for i in range(len(template)):
            check(item[i],template[i])

    if type(item) == types.DictionaryType:
        for key in template.keys():
            if not item.has_key(key):
                raise Error('data type error')
            check(item[key],template[key])

# Config files

if sys.platform == 'win32':
    config_dir = 'circle_config'
else:
    if os.environ.get('USER',os.environ.get('LOGNAME','root')) == 'root':
        config_dir = '/etc/circle'
    else:
        config_dir = os.path.join(os.environ['HOME'], '.circle')

if not os.path.isdir(config_dir):
    os.mkdir(config_dir)

def get_config(name, default):
    try:
        file_name = os.path.join(config_dir,name)
        os.chmod(file_name, 0600) # Only rw by user

        file = open(file_name,"rt")
        contents = file.read()
        file.close()

        try:
            return eval(contents)
        except:
            # Might be an old style config file
            return cPickle.loads(contents)
    except:
        return default

def set_config(name, value, fast=0):
    """ Write a config file.

            Writes to a new file then moves it, in order to ensure config files can
            not be corrupted."""
    file_name = os.path.join(config_dir,name)
    file = open(file_name+'.new','wt')

    try:
        os.chmod(file_name+'.new', 0600) # Only rw by user
    except OSError:
        pass

    #cPickle.dump(value, file, 1)
    if fast:
        file.write(repr(value))
    else:
        pprint.pprint(value, file)
    file.close()

    try:
        if os.path.isfile(file_name):
            os.unlink(file_name)
        os.rename(file_name+'.new',file_name)
    except OSError:
        pass

# Random bytes

def random_bytes(n):
    """ Make some random bytes, preferably cryptographic.

            This function is used to generate random bytes for challenges
            and for the node's address in the circle.  It would be
            preferable to have a urandom equivalent for windows, but
            alas..."""
    
    try:
        file = open('/dev/urandom','rb')
        result = file.read(n)
        file.close()
        return result
    except IOError:
        result = ''
        for i in range(n):
            result = result + chr(random.randint(0,255))
        return result

# Describe a number of bytes in readable form

def human_size(n):
    """ Return a logarithm scaled n for humans to understand information sizes."""

    # While KiB, etc are technically correct, more people are familiar with KB, etc
    
    if n < 1<<10:
        return "%d bytes" % n
    elif n < 1<<20:
        return "%.1fKB" % (n/float(1<<10))
    elif n < 1<<30:
        return "%.1fMB" % (n/float(1<<20))
    elif n < 1L<<40:
        return "%.1fGB" % (n/float(1<<30))
    else:
        return "%.1fTB" % (n/(2.0**40))

# Set terminal properties

def stdin_echo(on):
    """ Set character echo on the terminal on or off (used in chat text mode). """

    import termios

    # TERMIOS deprecated in 2.x series
    try:
        termios.ECHO
        TERMIOS = termios
    except:
        import TERMIOS

    new = termios.tcgetattr(0)
    if on:
        new[3] = new[3] | TERMIOS.ECHO | TERMIOS.ICANON
    else:
        new[3] = new[3] & ~TERMIOS.ECHO & ~TERMIOS.ICANON
    termios.tcsetattr(0, TERMIOS.TCSADRAIN, new)
 

# Quoting for the shell

def quote_for_shell(str):
    str = string.replace(str,'\\','\\\\')
    str = string.replace(str,'$','\\$')
    str = string.replace(str,'"','\\"')
    return '"'+string.replace(str,'`','\\`')+'"'
